//  ComputationalAssembler.java

import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;

import swarm.collections.ListImpl;
import swarm.collections.ListIndex;



/**
 * The ComputationalAssembler class instances make
 * computational processes; we have a computational assembler instance for
 * each unit
 *
 * @author Pietro Terna and Stefano T. Chiad
 */
public class ComputationalAssembler extends ComputationalAssemblerBasic{

    /**
     * the constructor for ComputationalAssembler
     *      */

    public ComputationalAssembler (Zone aZone, OrderDistiller od, ESFrameModelSwarm mo)
    {
	// Call the constructor for the parent class.
	super(aZone, od, mo);

    }

    /** drop the computational assembler */
    public void drop()
    {

	InformationRepository.addToHibernatedComputationalAssemblerList(this);
    }

    /** computational operations with code -12xx (to be kept in
     *  ComputationalAssembler and not to be transferred in
     *  ComputationalAssemblerBasic) <br><br>
     *
     */


     /** Agents are created and price is set for each of them.
      *  <br><br>
      *  computational operations with code -1201 (a code to set a price for the agent)
      *  <br><br>
      *  This code was written by Stefano T. Chiad
      *  <br><br>
      *  This computational code creates agents using the computational code c1101
      *  Then a random value, representing the price, is set in position 0,0
      *  of the unit matrix.
      *  Finally done is set to true.
      *
      */

    public void c1201() {

    	 Unit aUnit;
         int stratum;
	 int layer = 0;
	 float price = 0;
	 double min = 5;
	 double max = 20;
         float count;
         float counterMaxMissedTrades=5;


	 if(pendingComputationalSpecificationSet.getNumberOfMemoryMatrixesToBeUsed()!=1)
           {
             System.out.println("Code -1201 requires 1 matrix; " +
                                pendingComputationalSpecificationSet.getNumberOfMemoryMatrixesToBeUsed() +
                                " found in order # " +
                                pendingComputationalSpecificationSet.getOrderNumber());
             MyExit.exit(1);
           }

           /* the instruction checkMatrixNumber is very important to
              prevent an error. */
           checkMatrixNumber=false;
           aUnit = c1101();
           checkMatrixNumber=true;

           mm0 = (MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(0);
          layer = pendingComputationalSpecificationSet.getOrderLayer();

           // if the unit has been created, units created are counted for in matrix mm0
           if(aUnit != null)
           {
             count = mm0.getValue(layer,0,2);
             ++count;
             mm0.setValue(layer,0,2,count);
             System.out.println("The number of created units is : " + (int) count);
             // the matrix mm0 is printed on the console.
             mm0.print();
           }

           /* The "min" value is taken from the first general matrix (number 0)
              at position 1,0.     */
            min = (double) mm0.getValue(layer,1,0);
            System.out.println("Minimum value loaded from matrix 0 is : " + (int) min);
            /* The "max" value is taken from the first general matrix (number 0)
               at position 1,1.     */
            max = (double) mm0.getValue(layer,1,1);
            System.out.println("Maximum value loaded from matrix 0 is : " + (int) max);
            // Price is given using a function returning a random number.
            price = (float) Globals.env.uniformDblRand.getDoubleWithMin$withMax(min, max);
            System.out.println("The price set for this unit is: " + (int) price);

            // The "price" is put inside an agent's matrix.
            (aUnit.getMemoryMatrix()).setValue(layer,0,0,price);

            /* The counter for a firm maximum missed trades on the market.
               It will be used in another computational step.
               This counter will be decresead if the firm doesn't sell goods.
               When the counter will reach 0, the firm will get dropped.
               The counter is taken from general matrix 0 and then put inside
               an agent's matrix in positions 0,1 and 0,2.
            */

           stratum = (int) mm0.getValue(layer,0,1);
           System.out.println("The stratum for this unit is : " + (int) stratum);
            if (stratum == 0) {
              counterMaxMissedTrades = mm0.getValue(layer, 2, 0);
              (aUnit.getMemoryMatrix()).setValue(layer, 0, 1, counterMaxMissedTrades);
              (aUnit.getMemoryMatrix()).setValue(layer, 0, 2, counterMaxMissedTrades);
              System.out.println("The maximum missed trades possible is set to: " + (int) counterMaxMissedTrades);
            }

            done = true;

     }

     // end c1201

/** Firms and consumers exchange goods on the market.
      *  <br><br>
      *  computational operations with code -1202
      *  <br><br>
      *  This code was written by Stefano T. Chiad
      *  <br><br>
      *
      * This computational step compares two prices: one for a unit representing
      * a firm and one for a unit representing a consumer.
      * If the consumer buys a firm's goods, he/she remains on the market; otherwise
      * he/she exits from the market.
      * If a firm sells its goods it remains on the market. Otherwise a counter for
      * the maximum of missed trades is decresed by one.
      * Once the counter is zero the firm exits the market.
      * If a firm sells its goods the counter is restored to its default value.
      */

    public void c1202() {
      Order theCurrentOrder;
      Unit theUnitTheOrderIsIn, theUnitTheOrderComesFrom;
      int layer = 0;
      int firmNumber, consumerNumber;
      int check;
      float firmPrice = 0;
      float consumerPrice = 0;
      float counterMaxMissedTrades = 10;
      float counterActualMissedTrades = 10;
      float countDroppedFirms = 0;

      if(pendingComputationalSpecificationSet.getNumberOfMemoryMatrixesToBeUsed()!=2)
      {
        System.out.println("Code -1202 requires 2 matrixes; " +
                           pendingComputationalSpecificationSet.
                           getNumberOfMemoryMatrixesToBeUsed() +
                           " found in order # " + pendingComputationalSpecificationSet.
                           getOrderNumber());
        MyExit.exit(1);
      }

      // this code gets the two memory matrixes to be used later.
      mm0=(MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(0);
      mm1=(MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(1);

      layer=pendingComputationalSpecificationSet.getOrderLayer();
      theCurrentOrder=pendingComputationalSpecificationSet.getOrder();
      theUnitTheOrderIsIn=theCurrentOrder.getTheUnitTheOrderIsIn();
      theUnitTheOrderComesFrom=theCurrentOrder.getTheUnitTheOrderComesFrom();

      check = (int) (theUnitTheOrderIsIn.getMemoryMatrix()).getValue(layer,1,0);
      if (check == 1) {

          /* if the consumer has already exchaged in this cycle, the order is set to done */
          System.out.println("This consumer has already bought goods this time");
          done = true;

      }
      /* consumerPrice is imported from a consumer's memory matrix
         in layer 1 in position 0,0.
         Remember that consumers are in stratum 1.
       */
       consumerPrice = (theUnitTheOrderIsIn.getMemoryMatrix()).getValue(layer,0,0);
       System.out.println("The maximum price the consumer is willing to pay is: " + (int) consumerPrice);


       /* firmPrice is imported from a firm's memory matrix
          in layer 0 in position 0,0.
          Remember that firms are in stratum 0.
        */
       firmPrice = (theUnitTheOrderComesFrom.getMemoryMatrix()).getValue(layer,0,0);
       System.out.println("The minimum price for which the firm is willing to sell is: " + (int) firmPrice);

       /* counterMaxMissedTrades is the counter for the maximum
          missed trades allowed for a firm before it exits the market.
        */
        counterMaxMissedTrades = (theUnitTheOrderComesFrom.getMemoryMatrix()).getValue(layer,0,2);
        counterActualMissedTrades = (theUnitTheOrderComesFrom.getMemoryMatrix()).getValue(layer,0,1);
        System.out.println("The counter of possible missed trades for this firm is: " + (int) counterActualMissedTrades);

        if (consumerPrice <  firmPrice)
        {
            /* The consumer (theUnitTheOrderIsIn) exits from the market.
            Computational step c1102 drops the unit the order is in
            with a probability taken from position 0,0 of the first received matrix.
            Then it increases the value in position 0,1 of the second received
            matrix to count for the dropped units.
            So we need to put a row displacement of 2 and a column displacement
            of 1 to access the correct values in our general matrixes.
            */
            consumerNumber = theUnitTheOrderIsIn.getUnitNumber();
            System.out.println("Having not bought any goods, the consumer " + consumerNumber + " exits from the market");
            rd = 2;
            cd = 1;
            c1102();
            rd = 0;
            cd = 0;

            /* The "counterActualMissedTrades" is decreased by one.
            If it reaches 0 the firm is dropped.
            */
            counterActualMissedTrades = counterActualMissedTrades - 1;
            (theUnitTheOrderComesFrom.getMemoryMatrix()).setValue(layer,0,1,counterActualMissedTrades);
            if (counterActualMissedTrades == 0){
                /* firms who do not exchange for a long time are dropped
                and their number is accounted for in cell 2,2 in matrix 0.
                */
               countDroppedFirms = mm0.getValue(layer,2,2);
               mm0.setValue(layer,2,2,++countDroppedFirms);
               firmNumber = theUnitTheOrderComesFrom.getUnitNumber();
               System.out.println("Having not sold any goods, the firm " +  firmNumber + " exits from the market");
               System.out.println("The number of dropped firms is now:" + (int) countDroppedFirms);
               theUnitTheOrderComesFrom.getDroppable();

            }
          }

          /* If consumerPrice > firmPrice, then both the firm and the consumer
          remain on the market.
          In the firm's matrix the value "counterActualMissedTrades" is
          put back to its default value taken from "counterMaxMissedTrades".
          */
          else {
              /* If there is exchange of goods, we put a 1 in the consumer's matrix.
                 This means that that consumer has exchanged, so we proceed to another
                 recipe. */
            (theUnitTheOrderIsIn.getMemoryMatrix()).setValue(layer,1,0,1.0);
            counterActualMissedTrades = (theUnitTheOrderComesFrom.getMemoryMatrix()).getValue(layer, 0, 2);
            (theUnitTheOrderComesFrom.getMemoryMatrix()).setValue(layer,0,1,counterActualMissedTrades);
          }


          done = true;

       }

       // end c1202

       /** The check in consumers' matrixes is resetted to zero.
             *  <br><br>
             *  computational operations with code -1203
             *  <br><br>
             *  This code was written by Stefano T. Chiad
             *  <br><br>
             *
             * This computational step puts a zero at position 1,0 in the consumer's matrix.
             * Then sets done to true.
      */

       public void c1203() {
          Order theCurrentOrder;
          Unit theUnitTheOrderIsIn;

          if(pendingComputationalSpecificationSet.getNumberOfMemoryMatrixesToBeUsed()!=1)
          {
            System.out.println("Code -1203 requires 1 matrix; " +
                               pendingComputationalSpecificationSet.
                               getNumberOfMemoryMatrixesToBeUsed() +
                               " found in order # " + pendingComputationalSpecificationSet.
                               getOrderNumber());
            MyExit.exit(1);
          }

          theCurrentOrder=pendingComputationalSpecificationSet.getOrder();
          theUnitTheOrderIsIn=theCurrentOrder.getTheUnitTheOrderIsIn();
          (theUnitTheOrderIsIn.getMemoryMatrix()).setValue(layer,1,0, 0.0);

          done = true;
       }

       // end 1203

}



