//  ComputationalAssembler.java

import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;

import swarm.collections.ListImpl;
import swarm.collections.ListIndex;
import swarm.random.NormalDistImpl;


/**
 * The ComputationalAssembler class instances make
 * computational processes; we have a computational assembler instance for
 * each unit
 *
 * @author Pietro Terna and Stefano T. Chiad
 */
public class ComputationalAssembler extends ComputationalAssemblerBasic {


    /**
     * the constructor for ComputationalAssembler
     *      */

    public ComputationalAssembler (Zone aZone, OrderDistiller od, ESFrameModelSwarm mo)
    {
	// Call the constructor for the parent class.
	super(aZone, od, mo);

    }

    /** drop the computational assembler */
    public void drop()
    {

	InformationRepository.addToHibernatedComputationalAssemblerList(this);
    }


    /** computational operations with code -12xx (to be kept in
     *  ComputationalAssembler and not to be transferred in
     *  ComputationalAssemblerBasic) <br><br>
     *
     */


     /** Agents are created and price is set for each of them.
      *  <br><br>
      *  computational operations with code -1201 (a code to set a price for the agent)
      *  <br><br>
      *  This code was written by Stefano T. Chiad
      *  <br><br>
      *  This computational code creates agents using the computational code c1101
      *  Then a random value, representing the price, is set in position 0,0
      *  of the unit matrix.
      *  Finally done is set to true.
      *
      */

    public void c1201() {
         Unit aUnit;
         int stratum;
	 int layer = 0;
	 float price = 1;
	 double min = 5;
	 double max = 20;
         float count;
         float counterMaxMissedTrades=5;
         float consumerMaxMissedTrades=5;


	 if(pendingComputationalSpecificationSet.getNumberOfMemoryMatrixesToBeUsed()!=1)
           {
             System.out.println("Code -1201 requires 1 matrix; " +
                                pendingComputationalSpecificationSet.getNumberOfMemoryMatrixesToBeUsed() +
                                " found in order # " +
                                pendingComputationalSpecificationSet.getOrderNumber());
             MyExit.exit(1);
           }

           /* the instruction checkMatrixNumber is very important to
              prevent an error. */
           checkMatrixNumber=false;
           aUnit = c1101();
           checkMatrixNumber=true;

           mm0 = (MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(0);
           layer = pendingComputationalSpecificationSet.getOrderLayer();

           // if the unit has been created (it's not null) do as follows.
           if(aUnit != null)
           {
           // if the unit has been created, units created are counted for in matrix mm0
             count = mm0.getValue(layer, 0, 2);
             ++count;
             mm0.setValue(layer, 0, 2, count);
             System.out.println("The number of created units is : " + (int) count);

             /* The "min" value is taken from the first general matrix (number 0)
                at position 1,0.     */
             min = (double) mm0.getValue(layer, 1, 0);
             System.out.println("Minimum value loaded from matrix 0 is : " + (int) min);
             /* The "max" value is taken from the first general matrix (number 0)
                at position 1,1.     */
             max = (double) mm0.getValue(layer, 1, 1);
             System.out.println("Maximum value loaded from matrix 0 is : " + (int) max);
             // Price is given using a function returning a random number.
             price = (float) Globals.env.uniformDblRand.getDoubleWithMin$withMax(min, max);
             System.out.println("The price set for this unit is: " + (int) price);

             // The "price" is put inside an agent's matrix.
             (aUnit.getMemoryMatrix()).setValue(layer, 0, 0, price);

             /* the stratum on which the code is operating is taken from
                the general matrix. We will use it soon to differentiate between
                consumers and firms.
             */
             stratum = (int) mm0.getValue(layer, 0, 1);
             System.out.println("The stratum for this unit is : " + (int) stratum);

             if (stratum == 0) {
               /* if the agent is a firm do as follows.
               The counter for a firm maximum missed trades on the market.
               It will be used in another computational step.
               This counter will be decresead if the firm doesn't sell goods.
               When the counter will reach 0, the firm will get dropped.
               The counter is taken from general matrix 0 and then put inside
               an agent's matrix in positions 0,1 and 0,2.
               */
               counterMaxMissedTrades = mm0.getValue(layer, 2, 0);
               (aUnit.getMemoryMatrix()).setValue(layer, 0, 1, counterMaxMissedTrades);
               (aUnit.getMemoryMatrix()).setValue(layer, 0, 2, counterMaxMissedTrades);
               System.out.println("The maximum missed trades possible is set to: " +
                                  (int) counterMaxMissedTrades);
             }

             else if (stratum == 1) {
               /* if the agent is a consumer, then copy the price also at
                position 1,1 of a consumer's matrix.

                The counter for a consumer maximum missed trades on the market.
                It will be used in another computational step.
                This counter will be decresead if the consumer doesn't buy goods.
                When the counter will reach 0, the consumer will get dropped.
                The counter is taken from general matrix 1 at position 3,0 and
                then put inside an agent's matrix in positions 0,1 and 0,2.
                */
               consumerMaxMissedTrades = mm0.getValue(layer, 3, 0);
               (aUnit.getMemoryMatrix()).setValue(layer, 1, 1, price);
               (aUnit.getMemoryMatrix()).setValue(layer, 0, 1, consumerMaxMissedTrades);
               (aUnit.getMemoryMatrix()).setValue(layer, 0, 2, consumerMaxMissedTrades);
               /* if the agent is a firm copy the price at position 1,2 inside
                  the agent's matrix.
                */
               (aUnit.getMemoryMatrix()).setValue(layer, 1, 2, price);
             }

           }
            done = true;


     }

     // end c1201

/** Firms and consumers exchange goods on the market.
      *  <br><br>
      *  computational operations with code -1206
      *  <br><br>
      *  This code was written by Stefano T. Chiad
      *  <br><br>
      *
      * This computational step compares two prices: one for a unit representing
      * a firm and one for a unit representing a consumer.
      * Counters for both the firms and the consumers are used.
      * These counters are decreased by one if the agent doesn't trade.
      * Once the counter is zero the agent exits from the market.
      * If an agent trades any good the counter is restored to its default value.
      */

    public void c1206() {
      Order theCurrentOrder;
      Unit theUnitTheOrderIsIn, theUnitTheOrderComesFrom;
      int layer = 0;
      int firmNumber, consumerNumber;
      int check = 0;
      int checkFirm = 1;
      float firmPrice = 0;
      float consumerPrice = 0;
      float firmMaxMissedTrades = 10;
      float firmActualMissedTrades = 10;
      float consumerMaxMissedTrades =10;
      float consumerActualMissedTrades = 10;
      float countDroppedConsumers = 0;
      float countDroppedFirms = 0;
      float counterExchanges = 0;
      float oldAveragePrice = 0;
      float newAveragePrice = 0;


      if(pendingComputationalSpecificationSet.getNumberOfMemoryMatrixesToBeUsed()!=2)
      {
        System.out.println("Code -1206 requires 2 matrixes; " +
                           pendingComputationalSpecificationSet.
                           getNumberOfMemoryMatrixesToBeUsed() +
                           " found in order # " + pendingComputationalSpecificationSet.
                           getOrderNumber());
        MyExit.exit(1);
      }

      // this code gets the two memory matrixes to be used later.
      mm0=(MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(0);
      mm1=(MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(1);

      layer=pendingComputationalSpecificationSet.getOrderLayer();
      theCurrentOrder=pendingComputationalSpecificationSet.getOrder();
      theUnitTheOrderIsIn=theCurrentOrder.getTheUnitTheOrderIsIn();
      theUnitTheOrderComesFrom=theCurrentOrder.getTheUnitTheOrderComesFrom();

      check = (int) (theUnitTheOrderIsIn.getMemoryMatrix()).getValue(layer,1,0);

      if (check == 1) {

          /* if the consumer has already exchanged in this cycle, the order is set to done */
          System.out.println("This consumer has already bought goods this time");
          done = true;

      }

      /* consumerPrice is imported from a consumer's memory matrix
         in layer 1 in position 0,0.
         Remember that consumers are in stratum 1.
       */
      else if (check == 0) {
      consumerPrice = (theUnitTheOrderIsIn.getMemoryMatrix()).getValue(layer,0,0);

           /* The following code is a further check that the consumer price is set.
              If not the consumer gets dropped and the order is finished. */
           if (consumerPrice == 0) {
             theUnitTheOrderIsIn.getDroppable();
             done = true;

           }

           else if (consumerPrice != 0) {
             System.out.println("The maximum price the consumer is willing to pay is: " + (int) consumerPrice);

             /* firmPrice is imported from a firm's memory matrix in layer 0 in position 0,0.
                Remember that firms are in stratum 0.
              */
             firmPrice = (theUnitTheOrderComesFrom.getMemoryMatrix()).getValue(layer,0,0);
             System.out.println("The minimum price for which the firm is willing to sell is: " + (int) firmPrice);

             /* firmMaxMissedTrades is the counter for the maximum
                missed trades allowed for a firm before it exits the market.
              */
             firmMaxMissedTrades = (theUnitTheOrderComesFrom.getMemoryMatrix()).getValue(layer,0,2);
             firmActualMissedTrades = (theUnitTheOrderComesFrom.getMemoryMatrix()).getValue(layer,0,1);
             System.out.println("The counter of possible missed trades for this firm is: " + (int) firmActualMissedTrades);

             /* consumerMaxMissedTrades is the counter for the maximum
                missed trades allowed for a consumer before he/she exits from the market.
              */
             consumerMaxMissedTrades = (theUnitTheOrderIsIn.getMemoryMatrix()).getValue(layer,0,2);
             consumerActualMissedTrades = (theUnitTheOrderIsIn.getMemoryMatrix()).getValue(layer,0,1);
             System.out.println("The counter of possible missed trades for this consumer is: " + (int) consumerActualMissedTrades);



             if ( firmPrice > consumerPrice )
             {
                  /* The "consumerActualMissedTrades" is decreased by one.
                    If it reaches 0 the consumer is dropped.
                  */
                 consumerActualMissedTrades = consumerActualMissedTrades - 1;
                 (theUnitTheOrderIsIn.getMemoryMatrix()).setValue(layer,0,1,consumerActualMissedTrades);

                 if (consumerActualMissedTrades == 0){
                     /* consumers who do not exchange for a long time are dropped
                        and their number is accounted for in cell 2,2 in matrix 0.
                      */
                     countDroppedConsumers = mm1.getValue(layer,2,2);
                     mm1.setValue(layer,2,2,++countDroppedConsumers);
                     consumerNumber = theUnitTheOrderIsIn.getUnitNumber();
                     System.out.println("Having not bought any goods, the consumer " +  consumerNumber + " exits from the market");
                     System.out.println("The number of dropped consumers is now:" + (int) countDroppedConsumers);
                     theUnitTheOrderIsIn.getDroppable();
                   }

                 /* The "firmActualMissedTrades" is decreased by one.
                    If it reaches 0 the firm is dropped.
                  */
                 firmActualMissedTrades = firmActualMissedTrades - 1;
                 (theUnitTheOrderComesFrom.getMemoryMatrix()).setValue(layer,0,1,firmActualMissedTrades);

                 if (firmActualMissedTrades == 0){
                     /* firms who do not exchange for a long time are dropped
                        and their number is accounted for in cell 2,2 in matrix 0.
                      */
                     countDroppedFirms = mm0.getValue(layer,2,2);
                     mm0.setValue(layer,2,2,++countDroppedFirms);
                     firmNumber = theUnitTheOrderComesFrom.getUnitNumber();
                     System.out.println("Having not sold any goods, the firm " +  firmNumber + " exits from the market");
                     System.out.println("The number of dropped firms is now:" + (int) countDroppedFirms);
                     theUnitTheOrderComesFrom.getDroppable();
                   }
             }

             /* If consumerPrice > firmPrice, then both the firm and the consumer
                remain on the market.
                In the consumer's matrix the value "consumerActualMissedTrades" is
                put back to its default value taken from "consumerMaxMissedTrades".
                In the firm's matrix the value "firmActualMissedTrades" is
                put back to its default value taken from "firmMaxMissedTrades".
              */
             else {
                 /* If there is exchange of goods, we put a 1 at position 1,0
                    inside the agents' matrixes.
                 */
                 (theUnitTheOrderIsIn.getMemoryMatrix()).setValue(layer,1,0,1.0);
                 (theUnitTheOrderComesFrom.getMemoryMatrix()).setValue(layer,1,0,1.0);

                 /* we reset the counters of maximum missed trades before the agent
                    exits from the market. */
//                 consumerActualMissedTrades = (theUnitTheOrderIsIn.getMemoryMatrix()).getValue(layer, 0, 2);
                 (theUnitTheOrderIsIn.getMemoryMatrix()).setValue(layer,0,1,consumerMaxMissedTrades);

//                 firmActualMissedTrades = (theUnitTheOrderComesFrom.getMemoryMatrix()).getValue(layer, 0, 2);
                 (theUnitTheOrderComesFrom.getMemoryMatrix()).setValue(layer,0,1,firmMaxMissedTrades);
                 /*
                  the value checkFirm = 1 is put inside a firm's matrix at position 1,0.
                 */
                 (theUnitTheOrderComesFrom.getMemoryMatrix()).setValue(layer,1,0,checkFirm);

                 /* the number of exchanges done is registered in the general matrix 1
                    at position 1,2.
                 */
                 counterExchanges = (int) mm1.getValue(layer,1,2);
                 mm1.setValue(layer,1,2, ++counterExchanges);

                 /* the price of the exchange, supposed to be that of the firm,
                    is put inside the consumer's matrix at position 1,2.
                 */
                 (theUnitTheOrderIsIn.getMemoryMatrix()).setValue(layer,1,2,firmPrice);

                 /* the average price is registered and put inside general matrix 1
                    at position 4,1.
                  */
                 oldAveragePrice = mm1.getValue(layer,4,1);
                 newAveragePrice = ((counterExchanges*oldAveragePrice)+firmPrice)/(counterExchanges+1);
                 mm1.setValue(layer,4,1,newAveragePrice);

             }
       }
    }

    done = true;

}

       // end c1206

       /** The check in consumers' matrixes is resetted to zero.
             *  <br><br>
             *  computational operations with code -1203
             *  <br><br>
             *  This code was written by Stefano T. Chiad
             *  <br><br>
             *
             * This computational step puts a zero at position 1,0 in the consumer's matrix.
             * Then sets done to true.
      */

       public void c1203() {
          Order theCurrentOrder;
          Unit theUnitTheOrderIsIn;

          if(pendingComputationalSpecificationSet.getNumberOfMemoryMatrixesToBeUsed()!=1)
          {
            System.out.println("Code -1203 requires 1 matrix; " +
                               pendingComputationalSpecificationSet.
                               getNumberOfMemoryMatrixesToBeUsed() +
                               " found in order # " + pendingComputationalSpecificationSet.
                               getOrderNumber());
            MyExit.exit(1);
          }

          theCurrentOrder=pendingComputationalSpecificationSet.getOrder();
          theUnitTheOrderIsIn=theCurrentOrder.getTheUnitTheOrderIsIn();
          (theUnitTheOrderIsIn.getMemoryMatrix()).setValue(layer,1,0, 0.0);

          done = true;
       }

       // end 1203

       /** The consumers revise their prices.
             *  <br><br>
             *  computational operations with code -1207
             *  <br><br>
             *  This code was written by Stefano T. Chiad
             *  <br><br>
             *
             * This computational step lets the consumers revise their price.
             *
             * Then sets done to true.
      */

       public void c1207() {
              Order theCurrentOrder;
              Unit theUnitTheOrderIsIn;
              int layer = 0;
              int check = 2;
              double formerPrice = 0;
              double newPrice = 0;
              double variance = 0;
              double max = 0;
              double parameter = 1;

              if(pendingComputationalSpecificationSet.getNumberOfMemoryMatrixesToBeUsed()!=1)
              {
                System.out.println("Code -1207 requires 1 matrix; " +
                                   pendingComputationalSpecificationSet.
                                   getNumberOfMemoryMatrixesToBeUsed() +
                                   " found in order # " + pendingComputationalSpecificationSet.
                                   getOrderNumber());
                MyExit.exit(1);
              }

              mm0=(MemoryMatrix) pendingComputationalSpecificationSet.getMemoryMatrixAddress(0);

              theCurrentOrder = pendingComputationalSpecificationSet.getOrder();
              layer = pendingComputationalSpecificationSet.getOrderLayer();
              theUnitTheOrderIsIn = theCurrentOrder.getTheUnitTheOrderIsIn();
              NormalDistImpl myNormalDist = new NormalDistImpl (getZone(), Globals.env.randomGenerator);
              variance = (double) mm0.getValue(layer,4,2);

              check = (int) (theUnitTheOrderIsIn.getMemoryMatrix()).getValue(layer,1,0);
              /*
              the parameter is a factor that will "correct" the price obtained later
              from a normal distribution.
              It is obtained generating a random number between 0 and a value taken
              from the general matrix at position 3,2.
              */
              max = (double) mm0.getValue(layer,3,2);
              parameter = Globals.env.uniformDblRand.getDoubleWithMin$withMax(0, max);

              /* if the agent exchanged before it changes the price the agent had.
              */
              if (check == 1) {
              formerPrice = (theUnitTheOrderIsIn.getMemoryMatrix()).getValue(layer,0,0);
              newPrice = myNormalDist.getSampleWithMean$withVariance (formerPrice, variance);
              /*
              if the agent did not trade, the agent will revise
              probably downwards the price offered.
              */
              newPrice = newPrice * (1 - parameter) ;

              if (newPrice > 0) {
                (theUnitTheOrderIsIn.getMemoryMatrix()).setValue(layer, 0, 0, newPrice);
                System.out.println("The agent traded goods goods; his/her price was: " + formerPrice);
                System.out.println("The new price is: " + newPrice);
                }
              }

              /* if the agent did not exchange before, change the agent's price
              at position 0,0 in the agent's matrix.
              */
              if (check == 0) {
                formerPrice = (theUnitTheOrderIsIn.getMemoryMatrix()).getValue(layer, 0, 0);
                newPrice = myNormalDist.getSampleWithMean$withVariance(formerPrice, variance);
                /*
                 if the consumer did not exchange, he/she will revise
                 probably upwards the price offered.
                 */
                newPrice = newPrice * (1 + parameter);
                if (newPrice > 0) {
                  mm0.setValue(layer, 0, 0, newPrice);
                  System.out.println("The consumer didn't buy goods, his/her price was: " +
                                     formerPrice);
                  System.out.println("The new price is: " + newPrice);
                }
              }
              done = true;

           }

           // end 1207



}



